#!/usr/bin/perl -w
#######################################################
#
# Author: Polina Antipova aka Kitsune
#
# What it does:
# 	This script converts XLS files to XML format compatible with Testlink.
#
# More info at https://bitbucket.org/kitsuneo/xls2testlink-script/wiki
# Suggestion/bugs are also accepted at polina.antipova@gmail.com
#
# !! Script is very data sensible. Please, be careful with input XLS file !!
#
########################################################

use warnings;
use Spreadsheet::ParseExcel;
use File::Basename;
use feature qw(say);
use feature "switch";

sub log_and_print {
    print (scalar(localtime()), " ", @_, "\n");
    print LOGFILE (scalar(localtime()), " ", @_, "\n");
}

sub print_log {
	print LOGFILE (scalar(localtime()), " ", @_, "\n");
}

sub print_txt {
	print TXT (@_);
}

sub print_xml {
	print XML (@_);
}

# Source file and path are taken as script parameter
if (@ARGV) {
	$full_source_file_path = $ARGV[0];
	}

else {
	die (" ! Please, enter file name as script parameter. \n See README.txt for mor details");
	}

# Parse @ARGV, get path and file from script parameter
($source_file_name, $source_file_path, $source_file_suffix) = fileparse($full_source_file_path, qr/\.[^.]*/);	

# Define all files and paths
$source_file = "$source_file_name$source_file_suffix";

$logs_dir = "logs";
$converted_dir = "converted_files";

# Create folders for logs and converted files
chdir ($source_file_path);
mkdir ($logs_dir);
mkdir ($converted_dir);

# Output files get names after source file
$txt_file = ("$source_file_name\_parsed.txt");
$xml_file = ("$source_file_name\_resulted.xml");
$file_log = ("$source_file_name\_debug.log");

# Files + folders knocked into variables
$txt = "$converted_dir/$txt_file";
$xml = "$converted_dir/$xml_file";
$log = "$logs_dir/$file_log";

open (LOGFILE, "> $log") || die("Could not open file! $log");
open (TXT, "> $txt") || die("Could not open file! $txt");

log_and_print ("**************");
log_and_print ("Start working");
log_and_print ("STEP1: converting $source_file to TXT \n");

# see http://search.cpan.org/~jmcnamara/Spreadsheet-ParseExcel-0.59/lib/Spreadsheet/ParseExcel.pm#SYNOPSIS
# STEP1: The data from XLS file is stored in temp TXT file 
my $parser   = Spreadsheet::ParseExcel->new();
    my $workbook = $parser->parse($source_file);

    if ( !defined $workbook ) {
        die $parser->error(), ".\n";
    }
	
    for my $worksheet ( $workbook->worksheets() ) {

        my ( $row_min, $row_max ) = $worksheet->row_range();
        my ( $col_min, $col_max ) = $worksheet->col_range();

        for my $row ( $row_min .. $row_max ) {
            for my $col ( $col_min .. $col_max ) {

                my $cell = $worksheet->get_cell( $row, $col );
                next unless $cell;

               # Every cell is readed as row in .txt. Output format "Row, Col, value" 
			   $cell_unformatted = $cell->unformatted();	
			   $cell_unformatted =~ s/\n/<br>/g; 
			   
			   print_txt "$row;;$col;;", $cell_unformatted, "\n";
            }
        }
    }
	
close(TXT);	

log_and_print ("STEP1: OK \n Result: file $txt is created \n");	
log_and_print ("STEP2: converting $txt to XML \n");

# HERE THE KINGDOM OF CONVERTING BEGINS
open (XML, "> $xml") || die("Could not open file! $xml");
 
 # Read TXT to an array
	open (SOURCE, "$txt") || die ("Could not open file! $txt");
	@parsed_data = <SOURCE>;
	close (SOURCE);

	# Initialize auxiliary variables
	$source_row_counter = 0;
	$commented = "//";
	
	$header_of_xml = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<testsuite name = \"\">\n";
	$end_of_ts = "</testsuite>";
	
	# Print first row of XML file
	print_xml $header_of_xml;
	
 # Read every row from .txt array.	
 foreach $file_row (@parsed_data){	
 	
	chomp $file_row;
	
	# Cell is processed according to its positions and value
	@splited_row = split(/;;/, $file_row);
		
	$cell_row_position = $splited_row[0];
	$cell_col_position = $splited_row[1];
	$cell_value = $splited_row[2];
	
	print_log ("Debug: Current position: row - $cell_row_position, col - $cell_col_position");
	
	# First cell in excel file is expected to be mark of new test case (one test case - one excel row)
	# If cell is first, all test case variable are cleaned up
	if ($cell_col_position == 0){
		$test_case = "";
		$tc_name_tag = "";
		$tc_summary_tag = "";
		$tc_preconditions_tag = "";
		$tc_importance_tag = "";
		$tc_keywords_tag = "";
		$keywords_tagged = "";
		$tc_custom_fields_tag = "";
		$custom_fields_tagged = "";
		$test_case_actions = "";
		$test_case_expected_results = "";
		$tc_steps_tag = "";
		
		# TODO shift positions in arrays: in excel rows starts from 1, in parsed excel from 0
		print_log ("Debug: Start accumulate test case at row position $cell_row_position");
	}
	
	# Each value is checked if it is commented or not
	# Commented if cell is empty or first cell symbol is double slash "//"
	no warnings;
	$if_commented = substr $cell_value, 0, 2;
	
	# Skip cells from Row1 and Column A - reserved for Header and comments
	if (($cell_row_position == 0) | ($cell_col_position == 0)) {	
		print_log ("File: Skip \"Header or Comments\". Cell position: row - $cell_row_position, col - $cell_col_position");
	}
	
	# Skip empty cells
	# Check that 1st test case has a test suite name (!mandatory)
	elsif ($cell_value eq "") {	
		print_log ("File: Skip empty cell. \"$cell_value\", cell position: row - $cell_row_position, col - $cell_col_position");
		
		log_and_print ("! File: Test suite (row2) must be defined, otherwise .xml will be broken!") if (($cell_row_position ==1) && ($cell_col_position ==1));
		log_and_print ("! File: Test case name (row", $cell_row_position+1, ") must be defined, otherwise .xml will be broken!") if ($cell_col_position == 2);
	}
	
	# Skip cells started with "//"
	elsif ($if_commented eq $commented) {	
		print_log ("File: Skip empty cell. Cell begins with \"$if_commented\", cell position: row - $cell_row_position, col - $cell_col_position");
		log_and_print ("! File: Test suite (row2) must not be commented, otherwise .xml will be broken!") if (($cell_row_position == 1)&& ($cell_col_position ==1));
		log_and_print ("! File: Test case name (row", $cell_row_position+1, ") must not be commented, otherwise .xml will be broken!") if ($cell_col_position == 2);
	}
	
	# Accumulate test case record from all other values
	# Script use hard-coded values (columns numbers) and expects precise work with source file	
	else {	
		given ($cell_col_position) {
		
			# Block <test_suite>
			# Get test suite name, close previous test suite if it is not 1st test suite in the file
			when (1) {  my $ts_name = "$cell_value";
			
						my $ts_name_tag = "<testsuite name = \"$ts_name\">";	
						
						print_xml ($end_of_ts, "\n") if ($cell_row_position > 2);
						print_xml ($ts_name_tag, "\n");
					}
					
			# Block <test_case_name>	
			when (2) { my $tc_name = "$cell_value"; 
						$tc_name_tag = "<testcase name = \"$tc_name\">\n"
					}
					
			# Block <test_case_summary>		
			when (3) { my $tc_summary = "$cell_value"; 
						$tc_summary_tag = "<summary> <![CDATA[$tc_summary]]> </summary>\n";			
					}	
					
			# Block <test_case_preconditions>		
			when (4) { my $tc_preconditions = "$cell_value"; 
						$tc_preconditions_tag = "<preconditions> <![CDATA[$tc_preconditions]]> </preconditions>\n";
					}
					
			# Block <test_case_importance>		
			when (5) { my $tc_importance = "$cell_value";
						$tc_importance_tag = "<importance> <![CDATA[$tc_importance]]> </importance>\n"
					}
					
			# Block <keywords>
			# Kywords are expected have format "keyword1, keyword2, keyword3"	
			when (6) { my $tc_keywords = "$cell_value";
			
						@splited_keywords = split(/,/, $tc_keywords);
						
						foreach $keyword (@splited_keywords) {					
							my $keyword_tagged = "<keyword name=\"$keyword\"></keyword>";
							$keywords_tagged = "$keywords_tagged\n$keyword_tagged";
							$tc_keywords_tag = "<keywords> $keywords_tagged </keywords>\n";
						}
					}
					
			# Block <custom_fields>
			# Custom fields are expected to have format "name1 : value1, name2 : value2, name3 : value3"
			when (7) { my $tc_custom_fields = "$cell_value";
			
					@splited_custom_fields = split(/,/, $tc_custom_fields);
					
					foreach $custom_field(@splited_custom_fields) {
				
						@splited_custom_field = split (/:/, $custom_field);	
						
						my $cf_name = $splited_custom_field[0];
						my $cf_value = $splited_custom_field[1];
				
						my $cf_name_tagged = "<name><![CDATA[$cf_name]]></name>";
						my $cf_value_tagged = "<value><![CDATA[$cf_value]]></value>";												
						my $cf_tagged = "<custom_field>$cf_name_tagged\n$cf_value_tagged</custom_field>";
						$custom_fields_tagged = "$custom_fields_tagged\n$cf_tagged";
						}
						
				$tc_custom_fields_tag = "<custom_fields> $custom_fields_tagged </custom_fields>\n";
					
				}
			# Block <test_case_steps>
			when (8) { $test_case_actions = "$cell_value";}
			when (9) { $test_case_expected_results = "$cell_value";}		
		}
	
		# Block <steps> accumulates from columns [8]Steps - Actions [9]Steps - Expected result
		if (($test_case_actions ne "") | ($test_case_expected_results ne "")) {
			$tc_steps_tag = "<steps>
			<step>
			<step_number><![CDATA[1]]></step_number>
			<actions><![CDATA[$test_case_actions]]></actions>
			<expectedresults><![CDATA[$test_case_expected_results]]></expectedresults>
			</step>
			</steps>\n";
		}
		
		# Final accumulated test case
		$test_case = "$tc_name_tag $tc_summary_tag $tc_preconditions_tag $tc_importance_tag $tc_keywords_tag $tc_custom_fields_tag $tc_steps_tag	</testcase>\n";	
	}
	
	# When cell is last in the excel row - call to print test case to file
	if ($cell_col_position == 9) {
		$source_row_counter++;
		
		print_log ("File: Print accumulated test case at row position $cell_row_position");
		print_xml $test_case;		
	}
 }
  
 # Close XML file - double </testsuite>
 print_xml ($end_of_ts, "\n", $end_of_ts);
  
 # deduct one row from row counter (reserved for headers) for get number of processed Excel test cases
 $source_row_counter = $source_row_counter -1; 
 
 log_and_print ("STEP2: OK \n Result: file $xml is created \n");	
 log_and_print ("Processed $source_row_counter rows from $source_file \n See debug log for more details.\n\n");
 
close(XML);
close(LOGFILE);